import numpy as np
from matplotlib import colors
import matplotlib.pyplot as plt
import geone
import geone.covModel as gcm
import geone.imgplot3d as imgplt3
import pyvista as pv
import sys
import os
import flopy as fp
import pandas as pd

# some functions
def mask_below_unit(T1, unit, iu=0):
    u = T1.get_unit(unit)
    
    mask_tot = np.zeros([T1.nz, T1.ny, T1.nx])
    mask_unit = T1.unit_mask(u.name, iu=iu)
    mask_tot[mask_unit.astype(bool)] = 1

    for unit_to_compare in T1.get_all_units():
        if unit_to_compare > u:
            mask_unit = T1.unit_mask(unit_to_compare.name, iu=iu)
            mask_tot[mask_unit.astype(bool)] = 1
    
    return mask_tot

def cellidBD(idomain, layer=0):   
    
    """
    extract the cellids at the boundary of the domain at a given layer
    idomain : 3D array, idomain array which determine if a cell is active or not (1 active, 0 inactive)
    layer : int, layer on which the boundary cells are extract
    """
    lst_cellBD=[]

    for irow in range(idomain.shape[1]):
        for icol in range(idomain.shape[2]):
            if idomain[layer][irow,icol]==1:
                #check neighbours
                if np.sum(idomain[layer][irow-1:irow+2,icol-1:icol+2]==1) < 8:
                    lst_cellBD.append((layer,irow,icol))
    return lst_cellBD

def check_thk(top, botm):
    """
    check if all the cells in a modflow model (given top and botm array) which have a thickness <= 0 for and this for each layer
    input : top (the top surface) and botm (botom of each layer)
    output : lst of bool (false mean everything's okay in that specific layer !)
    """
    nlay = botm.shape[0]
    bol_lst=[]
    bol_lst.append(not ((top-botm[0])<=0).any())
    for ilay in range(nlay-1):
        bol_lst.append(not ((botm[ilay]-botm[ilay+1])<=0).any())
    return bol_lst

def array2cellids(array, idomain):

    """
    Convert an array to a list of cellids
    array : 3D array, array to convert
    idomain : 3D array, idomain array which determine if a cell is active or not (1 active, 0 inactive)
    """
    cellids = []

    idomain_copy = idomain.copy()
    idomain_copy[idomain == -1] = 0
    b = array & idomain_copy.astype(bool)
    for ilay in range(b.shape[0]):
        for irow in range(b.shape[1]):
            for icol in range(b.shape[2]):
                if b[ilay, irow, icol]:
                    cellids.append((ilay, irow, icol))
    return cellids

def plot_particle_facies_sequence(arch_table, df, plot_time=False, plot_distance=False, proportions=False):

    if plot_time and plot_distance:
        fig, ax = plt.subplots(2,1, figsize=(10, 1.5), dpi=200)
        axi = ax[0]
    elif plot_time or plot_distance:
        fig, axi = plt.subplots(1,1, figsize=(10, 0.5), dpi=200)
    plt.subplots_adjust(hspace=1.5)

    if proportions:
        colors_fa = []
        for col in df.columns:
            if col.split("_")[0] == "facies":
                id_fa = int(col.split("_")[-1])
                color_fa = archpy_flow.T1.get_facies_obj(ID=id_fa, type="ID").c
                colors_fa.append(color_fa)

    if plot_time:

        if proportions:
            df.set_index("time").iloc[:, -len(colors_fa):].plot(color=colors_fa, legend=False, ax=ax[0])
            axi.set_ylabel("Proportion")
            axi.set_xlabel("time [days]")
            axi.set_ylim(-.1, 1.1)

        else:
            dt = df["dt"]
            for i, (facies, time) in enumerate(zip(df["facies"], df["time"])):
                if i > 0:
                    axi.barh(0, dt[i], left=df["time"].loc[i], color=arch_table.get_facies_obj(ID=facies, type="ID").c)
                else:
                    axi.barh(0, dt[i], left=0, color=arch_table.get_facies_obj(ID=facies, type="ID").c, label=arch_table.get_facies_obj(ID=facies, type="ID").name)
            
            axi.set_xlim(0, df["time"].iloc[-1])
            axi.set_xlabel("Time (days)")
            axi.set_yticks([])

        if plot_distance:
            axi = ax[1]
    
    # plot facies function of the distance traveled
    if plot_distance:
        all_dist = df["distance"].values
        all_cum_dist = df["cum_distance"].values

        if proportions:
            df.set_index("cum_distance").iloc[:, -len(colors_fa):].plot(color=colors_fa, legend=False, ax=ax[1])
            ax[1].set_ylabel("Proportion")
            ax[1].set_xlabel("Distance [m]")
            ax[1].set_ylim(-.1, 1.1)

        else:

            for i in range(int(len(df["facies"]) - 1)):
                facies = df["facies"].iloc[i]
                width = all_dist[i+1]
                distance = all_cum_dist[i]
                axi.barh(0, width, left=distance, color=arch_table.get_facies_obj(ID=facies, type="ID").c)

        axi.set_xlim(0, all_cum_dist[-1])
        axi.set_xlabel("Distance (m)")
        axi.set_yticks([])


def get_nodes(locs, nx, ny):
    nodes = []
    for k, i, j in locs:
        nodes.append(k * nx * ny + i * nx + j)
    return nodes

def get_locs(nodes, nx, ny):
    locs = []
    for node in nodes:
        k = node // (nx * ny)
        i = (node - k * nx * ny) // nx
        j = node - k * nx * ny - i * nx
        locs.append((k, i, j))
    return locs

## archpy to modflow class ##
class archpy2modflow:

    """
    Class to convert an ArchPy table to a MODFLOW 6 model

    Parameters
    ----------
    T1 : Arch_table
        ArchPy table to convert
    sim_name : str
        name of the simulation
    model_dir : str
        directory where the model will be saved
    model_name : str
        name of the model
    exe_name : str
        path to the mf6 executable
    """

    def __init__(self, T1, sim_name="sim_test", model_dir="workspace", model_name="test", exe_name="mf6"):
        self.T1 = T1
        self.sim_name = sim_name
        self.model_dir = model_dir
        self.model_name = model_name
        self.exe_name = exe_name
        self.sim = None
        self.grid_mode = None
        self.layers_names = None
        self.list_active_cells = None
        self.mp = None
        self.factor_x = None
        self.factor_y = None
        self.factor_z = None

    def create_sim(self, grid_mode="archpy", iu=0, factor_x=None, factor_y=None, factor_z=None, unit_limit=None):

        """
        Create a modflow simulation from an ArchPy table
        
        Parameters
        ----------
        grid_mode : str
            "archpy" : use the grid defined in the ArchPy table
            "layers" : use the surfaces of each unit to define the grid
            "new_resolution" : use factors to change the resolution of the grid
            In this case, factor_x, factor_y and factor_z must be provided
        iu : int
            index of the unit to use when grid_mode is "layers"
        factor_x : float
            factor to change the resolution of the grid in the x direction. e.g. 2 means that the resolution will be divided by 2
        factor_y : float
            factor to change the resolution of the grid in the y direction.
        factor_z : float    
            factor to change the resolution of the grid in the z direction.
        unit_limit: unit name
            unit under which cells are considered as inactive
        """

        sim = fp.mf6.MFSimulation(sim_name=self.sim_name, version='mf6', exe_name=self.exe_name, 
                         sim_ws=self.model_dir)
        gwf = fp.mf6.ModflowGwf(sim, modelname=self.model_name,
                                model_nam_file='{}.nam'.format(self.model_name))

        #grid
        nlay, nrow, ncol = self.T1.get_nz(), self.T1.get_ny(), self.T1.get_nx()
        delr, delc = self.T1.get_sx(), self.T1.get_sy()
        xoff, yoff = self.T1.get_ox(), self.T1.get_oy()

        if grid_mode == "archpy":
            top = np.ones((nrow, ncol)) * self.T1.get_zg()[-1]
            botm = np.ones((nlay, nrow, ncol)) * self.T1.get_zg()[:-1].reshape(-1, 1, 1)
            botm = np.flip(np.flipud(botm), axis=1)  # flip the array to have the same orientation as the ArchPy table
            idomain = np.flip(np.flipud(self.T1.get_mask().astype(int)), axis=1)  # flip the array to have the same orientation as the ArchPy table

            # inactive cells below unit limit
            if unit_limit is not None:
                mask = mask_below_unit(self.T1, unit_limit, iu=iu)
                mask = np.flip(np.flipud(mask), axis=1)  # flip the array to have the same orientation as the ArchPy table
                idomain[mask == 1] = 0

        elif grid_mode == "layers":

            def list_unit_below_unit(T1, unit):
                u = T1.get_unit(unit)
                units = []
                for unit_to_compare in T1.get_all_units():
                    if unit_to_compare > u:
                        units.append(unit_to_compare)
                return units

            # determine which units will be inactive
            if unit_limit is not None:
                units = list_unit_below_unit(self.T1, unit_limit)
                n_units_removed = len(units) + 1
            else:
                n_units_removed = None

            # get surfaces of each unit
            top = self.T1.get_surface(typ="top")[0][0, iu]
            top = np.flip(top, axis=0)
            botm = self.T1.get_surface(typ="bot")[0][:, iu]
            botm = np.flip(botm, axis=1)
            layers_names = self.T1.get_surface(typ="bot")[1]
            self.layers_names = layers_names
            nlay = botm.shape[0]

            # define idomain (1 if thickness > 0, 0 if nan, -1 if thickness = 0)
            idomain = np.ones((nlay, nrow, ncol))
            thicknesses = -np.diff(np.vstack([top.reshape(-1, nrow, ncol), botm]), axis=0)
            idomain[thicknesses == 0] = -1
            idomain[np.isnan(thicknesses)] = 0

            # inactive cells below unit limit
            if n_units_removed is not None:
                idomain[-n_units_removed:] = 0

            # adapt botm in order that each layer has a thickness > 0 
            for i in range(-1, nlay-1):
                if i == -1:
                    s1 = top
                else:
                    s1 = botm[i]
                s2 = botm[i+1]
                mask = s1 == s2
                s1[mask] += 1e-2

                # 2nd loop over previous layers to ensure that the thickness is > 0
                for o in range(i, -1, -1):
                    s2 = botm[o]
                    if o == 0:
                        s1 = top
                    else:
                        s1 = botm[o-1]
                    mask = s1 <= s2
                    s1[mask] = s2[mask] + 1e-2

        elif grid_mode == "new_resolution":
            assert factor_x is not None, "factor_x must be provided"
            assert factor_y is not None, "factor_y must be provided"
            assert factor_z is not None, "factor_z must be provided"
            assert nrow % factor_y == 0, "nrow must be divisible by factor_y"
            assert ncol % factor_x == 0, "ncol must be divisible by factor_x"
            assert nlay % factor_z == 0, "nlay must be divisible by factor_z"
            nrow = int(nrow / factor_y)
            ncol = int(ncol / factor_x)
            nlay = int(nlay / factor_z)
            delr = delr * factor_x
            delc = delc * factor_y
            top = np.ones((nrow, ncol)) * self.T1.get_zg()[-1]
            botm = np.ones((nlay, nrow, ncol)) * self.T1.get_zg()[::-factor_z][1:].reshape(-1, 1, 1)
            # botm = np.flip(np.flipud(botm), axis=1)  # flip the array to have the same orientation as the ArchPy table
            
            # how to define idomain ?
            idomain = np.zeros((nlay, nrow, ncol))
            mask_org = self.T1.get_mask().astype(int)

            # modify mask_org to remove the inactive cells
            if unit_limit is not None:
                mask = mask_below_unit(self.T1, unit_limit, iu=iu)
                mask_org[mask == 1] = 0

            for ilay in range(0, self.T1.get_nz(), factor_z):
                for irow in range(0, self.T1.get_ny(), factor_y):
                    for icol in range(0, self.T1.get_nx(), factor_x):
                        mask = mask_org[ilay:ilay+factor_z, irow:irow+factor_y, icol:icol+factor_x]
                        if mask.mean() >= 0.5:
                            idomain[ilay//factor_z, irow//factor_y, icol//factor_x] = 1
            
            idomain = np.flip(np.flipud(idomain), axis=1)  # flip the array to have the same orientation as the ArchPy table

            self.factor_x = factor_x
            self.factor_y = factor_y
            self.factor_z = factor_z
            
        else:
            raise ValueError("grid_mode must be one of 'archpy', 'layers' or 'new_resolution'")

        # save grid mode
        self.grid_mode = grid_mode
        
        assert (np.array(check_thk(top, botm))).all(), "Error in the processing of the surfaces, some cells have a thickness < 0"

        dis = fp.mf6.ModflowGwfdis(gwf, nlay=nlay, nrow=nrow, ncol=ncol,
                                    delr=delr, delc=delc,
                                    top=top, botm=botm,
                                    xorigin=xoff, yorigin=yoff, 
                                    idomain=idomain)

        perioddata = [(1, 1, 1.0)]
        tdis = fp.mf6.ModflowTdis(sim, time_units='SECONDS',perioddata=perioddata)

        ims = fp.mf6.ModflowIms(sim, complexity="simple")

        #Initial condition
        ic   = fp.mf6.ModflowGwfic(gwf, strt=1)

        # output control
        oc   = fp.mf6.ModflowGwfoc(gwf,budget_filerecord='{}.cbc'.format(self.model_name),
                                    head_filerecord='{}.hds'.format(self.model_name),
                                    saverecord=[('HEAD', 'LAST'),
                                                ('BUDGET', 'LAST')],
                                    printrecord=[('BUDGET', 'ALL')])
        
        # npf package
        # empty package
        npf = fp.mf6.ModflowGwfnpf(gwf, icelltype=0, k=1, save_flows=True, save_saturation=True)

        self.sim = sim
        print("Simulation created")
        print("To retrieve the simulation, use the get_sim() method")
    
    def set_k(self, k_key="K",
              iu=0, ifa=0, ip=0,
              log=False, k=None, k22=None, k33=None, k_average_method="arithmetic", 
              upscaling_method="simplified_renormalization"):

        """
        Set the hydraulic conductivity for a specific facies
        """

        # write a function to get proportion of each value in the array
        def get_proportion(arr):
            unique, counts = np.unique(arr, return_counts=True)
            return dict(zip(unique, counts/np.sum(counts)))

        # remove the npf package if it already exists
        gwf = self.get_gwf()
        gwf.remove_package("npf")

        if k is None:
            grid_mode = self.grid_mode
            if grid_mode == "archpy":
                new_k22 = None
                new_k33 = None
                k = self.T1.get_prop(k_key)[iu, ifa, ip]
                k = np.flip(np.flipud(k), axis=1)  # flip the array to have the same orientation as the ArchPy table
                if log:
                    new_k = 10**k
                else:
                    new_k = k

            elif grid_mode == "layers":

                new_k22 = None
                nrow, ncol, nlay = gwf.modelgrid.nrow, gwf.modelgrid.ncol, gwf.modelgrid.nlay

                # initialize new_k and new_k33
                kh = self.T1.get_prop(k_key)[iu, ifa, ip] 
                new_k = np.ones((nlay, nrow, ncol))

                # initialize variable for facies upscaling
                facies_arr = self.T1.get_facies(iu, ifa, all_data=False)
                upscaled_facies = {}

                for ifa in np.unique(facies_arr):
                    upscaled_facies[ifa] = np.zeros((facies_arr.shape[0], facies_arr.shape[1], facies_arr.shape[2]))

                if k_average_method == "anisotropic":
                    new_k33 = np.ones((nlay, nrow, ncol))
                else:
                    new_k33 = None
                layers = self.layers_names
                mask_units = [self.T1.unit_mask(l).astype(bool) for l in layers]

                for irow in range(nrow):
                    for icol in range(ncol):
                        for ilay in range(nlay):
                            mask_unit = mask_units[ilay]
                            if k_average_method == "arithmetic":
                                new_k[ilay, irow, icol] = np.mean(kh[:, irow, icol][mask_unit[:, irow, icol]])
                            elif k_average_method == "harmonic":
                                new_k[ilay, irow, icol] = 1 / np.mean(1 / kh[:, irow, icol][mask_unit[:, irow, icol]])
                            elif k_average_method == "anisotropic":
                                new_k[ilay, irow, icol] = np.mean(kh[:, irow, icol][mask_unit[:, irow, icol]])
                                new_k33[ilay, irow, icol] = 1 / np.mean(1 / kh[:, irow, icol][mask_unit[:, irow, icol]])
                            else:
                                raise ValueError("k_average_method must be one of 'arithmetic' or 'harmonic'")
                            
                            # facies upscaling
                            arr = facies_arr[:, irow, icol][mask_unit[:, irow, icol]]  # array of facies values in the unit
                            prop = get_proportion(arr)
                            for ifa in np.unique(arr):
                                upscaled_facies[ifa][:, irow, icol][mask_unit[:, irow, icol]] = prop[ifa]
                
                # save upscaled facies
                self.upscaled_facies = upscaled_facies

                # fill nan values with the mean of the layer
                for ilay in range(nlay):
                    mask = np.isnan(new_k[ilay])
                    new_k[ilay][mask] = np.nanmean(new_k[ilay])

                new_k = np.flip(new_k, axis=1)  # we have to flip in order to match modflow grid

                if k_average_method == "anisotropic":
                    for ilay in range(nlay):
                        mask = np.isnan(new_k33[ilay])
                        new_k33[ilay][mask] = np.nanmean(new_k33[ilay])
                    
                    new_k33 = np.flip(new_k33, axis=1)

                if log:
                    new_k = 10**new_k
                    if k_average_method == "anisotropic":
                        new_k33 = 10**new_k33

            elif grid_mode == "new_resolution":

                from ArchPy.uppy import upscale_k
                dx, dy, dz = self.T1.get_sx(), self.T1.get_sy(), self.T1.get_sz()
                
                factor_x = self.factor_x
                factor_y = self.factor_y
                factor_z = self.factor_z
                
                field = self.T1.get_prop(k_key)[iu, ifa, ip]
                field = np.flip(np.flipud(field), axis=1)  # flip the array to have the same orientation as the ArchPy table
                
                if upscaling_method == "standard_renormalization_center":
                    field_kxx, field_kyy, field_kzz = upscale_k(field, method="standard_renormalization", dx=dx, dy=dy, dz=dz, factor_x=factor_x, factor_y=factor_y, factor_z=factor_z, scheme="center")
                elif upscaling_method == "standard_renormalization_direct":
                    field_kxx, field_kyy, field_kzz = upscale_k(field, method="standard_renormalization", dx=dx, dy=dy, dz=dz, factor_x=factor_x, factor_y=factor_y, factor_z=factor_z, scheme="direct")
                elif upscaling_method == "tensorial_renormalization":
                    field_kxx, field_kyy, field_kzz = upscale_k(field, method="tensorial_renormalization", dx=dx, dy=dy, dz=dz, factor_x=factor_x, factor_y=factor_y, factor_z=factor_z)
                elif upscaling_method == "simplified_renormalization":
                    field_kxx, field_kyy, field_kzz = upscale_k(field, method="simplified_renormalization", dx=dx, dy=dy, dz=dz, factor_x=factor_x, factor_y=factor_y, factor_z=factor_z)
                
                new_k = field_kxx
                new_k22 = field_kyy
                new_k33 = field_kzz
                
                # fill nan values
                new_k[np.isnan(new_k)] = np.nanmean(new_k)
                new_k22[np.isnan(new_k22)] = np.nanmean(new_k22)
                new_k33[np.isnan(new_k33)] = np.nanmean(new_k33)

                if log:
                    new_k = 10**new_k
                    new_k22 = 10**new_k22
                    new_k33 = 10**new_k33

                # facies upscaling
                facies_arr = self.T1.get_facies(iu, ifa, all_data=False)
                upscaled_facies = {}
                for ifa in np.unique(facies_arr):
                    upscaled_facies[ifa] = np.zeros((facies_arr.shape[0], facies_arr.shape[1], facies_arr.shape[2]))
                
                for ilay in range(0, self.T1.get_nz(), factor_z):
                    for irow in range(0, self.T1.get_ny(), factor_y):
                        for icol in range(0, self.T1.get_nx(), factor_x):
                            mask_unit = facies_arr[ilay:ilay+factor_z, irow:irow+factor_y, icol:icol+factor_x]
                            arr = mask_unit.flatten()
                            prop = get_proportion(arr)
                            for ifa in np.unique(arr):
                                upscaled_facies[ifa][ilay:ilay+factor_z, irow:irow+factor_y, icol:icol+factor_x] = prop[ifa]

                self.upscaled_facies = upscaled_facies
        else:
            new_k = k
            new_k22 = k22
            new_k33 = k33

        # new_k = np.flip(new_k, axis=1)  # we have to flip in order to match modflow grid
        npf = fp.mf6.ModflowGwfnpf(gwf, icelltype=0, k=new_k, k22=new_k22, k33=new_k33, save_flows=True, save_specific_discharge=True, save_saturation=True)
        npf.write()


    def set_strt(self, heads=None):
        """
        Set the starting heads
        """
        gwf = self.get_gwf()
        gwf.remove_package("ic")
        ic = fp.mf6.ModflowGwfic(gwf, strt=heads)
        ic.write()

    def get_list_active_cells(self):

        if self.list_active_cells is None:
            gwf = self.get_gwf()
            idomain = gwf.dis.idomain.array
            list_active_cells = []
            for ilay in range(idomain.shape[0]):
                for irow in range(idomain.shape[1]):
                    for icol in range(idomain.shape[2]):
                        if idomain[ilay, irow, icol] == 1:
                            list_active_cells.append((ilay, irow, icol))
            self.list_active_cells = list_active_cells
            
        return self.list_active_cells

    # get functions
    def get_sim(self):
        assert self.sim is not None, "You need to create the simulation first"
        return self.sim
    
    def get_gwf(self):
        assert self.sim is not None, "You need to create the simulation first"
        return self.sim.get_model()

    # outputs
    def get_heads(self, kstpkper=(0, 0)):
        """
        Get the heads of the simulation
        """
        gwf = self.get_gwf()
        head = gwf.output.head().get_data(kstpkper=kstpkper)
        return head
    
    # plots
    def plot_3D_heads(self, kstpkper=(0, 0), ax=None, **kwargs):

        """
        Plot the heads of the simulation
        """
        assert self.grid_mode == "archpy", "This function only works with the 'archpy' grid mode"
        
        head = self.get_heads(kstpkper=kstpkper)
        head[head == 1e30] = np.nan

        self.T1.plot_arr(np.flipud(np.flip(head, axis=1)), "head", 2)

    def mp_create(self, mpexe, trackdir="forward",
                  locs=None, rowcelldivisions=1, columncelldivisions=1, layercelldivisions=1,
                  list_p_coords=None):
        """
        Create a modpath simulation from an ArchPy table

        Parameters
        ----------
        mpexe : str
            path to the modpath executable
        trackdir : str
            direction of tracking
        locs : list of tuples
            list of cells to put particles. List of int where values correspond to the indices of the cells.
            Each cell in locs will be then spatially divided and will output rowcelldivisions * columncelldivisions * layercelldivisions particles
        rowcelldivisions : int
            number of row divisions. If 1, no division, if 2, divide each cell in 2 parts, etc.
        columncelldivisions : int
            number of column divisions
        layercelldivisions : int
            number of layer divisions
        list_p_coords : list of tuples
            list of particles coordinates. Each tuple must have 3 values (xp, yp, zp) corresponding to the coordinates of the particle
        """

        gwf = self.get_gwf()
        workspace = self.model_dir
        model_name = gwf.name
        mpnamf = f"{model_name}_mp_forward"
        
        if locs is not None:
            nodes = get_nodes(locs, gwf.modelgrid.ncol, gwf.modelgrid.nrow)
            if len(nodes) == 0:
                print("No particles to track")
                return
            elif len(nodes) == 1:
                nodes = nodes[0]
            mp = fp.modpath.Modpath7.create_mp7(
                                                modelname=mpnamf,
                                                trackdir=trackdir,
                                                flowmodel=gwf,
                                                model_ws=workspace,
                                                rowcelldivisions=rowcelldivisions,
                                                columncelldivisions=columncelldivisions,
                                                layercelldivisions=layercelldivisions,
                                                nodes = nodes,
                                                exe_name=mpexe,
                                            )
        else:
            if list_p_coords is not None:
                # write a function to find the modlfow cellids as well as localx, localy and localz from a coordinate
                from shapely.geometry import Point, MultiPoint

                grid = self.get_gwf().modelgrid
                ix = fp.utils.gridintersect.GridIntersect(mfgrid=grid)

                list_p = []
                list_cellids = []
                for pi in list_p_coords:
                    p1 = Point(pi)
                    list_cellids.append(ix.intersect(p1).cellids[0])

                cellids = np.array([cids for cids in list_cellids])

                # multp = MultiPoint(list_p)

                # cellids = ix.intersect(multp).cellids
                # cellids = np.array([np.array(cids) for cids in cellids])
                cellids[:, 0] += 1

                l = []
                for i in range(len(cellids)):
                    cid = cellids[i]
                    exem_i = list_p_coords[i]
            
                    local_dx = (grid.xvertices[cid[1], cid[2]+1] - grid.xvertices[cid[1], cid[2]]) / 2
                    local_dy = - (grid.yvertices[cid[1]+1, cid[2]] - grid.yvertices[cid[1], cid[2]]) / 2
                    # local_dz = - (grid.zcellcenters[cid[0]+1, cid[1], cid[2]] - grid.zcellcenters[cid[0], cid[1], cid[2]]) / 2
                    local_dz = (grid.botm[cid[0]-1, cid[1], cid[2]] - grid.botm[cid[0], cid[1], cid[2]]) / 2
                    localx = (exem_i[0] - (grid.xcellcenters[cid[1], cid[2]] - local_dx)) / (2*local_dx)
                    localy = (exem_i[1] - (grid.ycellcenters[cid[1], cid[2]] - local_dy)) / (2*local_dy)
                    localz = (exem_i[2] - (grid.zcellcenters[cid[0], cid[1], cid[2]] - local_dz)) / (2*local_dz)

                    p1 = fp.modpath.mp7particledata.ParticleData([tuple(cid)], structured=True, localx=localx, localy=localy, localz=localz)
                    pg = fp.modpath.mp7particlegroup.ParticleGroup(particledata=p1)

                    l.append(pg)
                    
                # create a modpath simulation
                mp = fp.modpath.Modpath7(
                    modelname=mpnamf,
                    flowmodel=gwf,
                    exe_name=mpexe,
                    model_ws=workspace,
                    verbose=0,
                )

                # basic package
                fp.modpath.Modpath7Bas(mp)

                fp.modpath.Modpath7Sim(
                    mp,
                    simulationtype="combined",
                    trackingdirection=trackdir,
                    weaksinkoption="pass_through",
                    weaksourceoption="pass_through",
                    referencetime=0.0,
                    stoptimeoption="extend",
                    particlegroups=l,
                )

        self.mp = mp  # save the modpath object
        self.mpnamf = mpnamf  # save the name of the modpath file

    def prt_create(self, prt_name="test", workspace="./", trackdir="forward", list_p_coords=None):
        
        """
        Create a particle tracking simulation from a list of coordinates

        Parameters
        ----------
        prt_name : str
            name of the particle tracking simulation
        workspace : str
            path to the workspace where the simulation will be saved
        trackdir : str
            tracking direction. Can be "forward" or "backward"
        list_p_coords : list of tuples
            list of particles coordinates. Each tuple must have 3 values (xp, yp, zp) corresponding to the coordinates of the particle
        """

        sim_prt = fp.mf6.MFSimulation(sim_name=prt_name, exe_name=self.exe_name, sim_ws=workspace, version="mf6")

        tdis_perioddata = self.get_sim().tdis.perioddata.array
        tdis = fp.mf6.ModflowTdis(sim_prt, pname="tdis", time_units="DAYS", nper=len(tdis_perioddata), perioddata=tdis_perioddata)

        # Create PRT model
        prt = fp.mf6.ModflowPrt(sim_prt, modelname=prt_name, model_nam_file="{}.nam".format(prt_name))

        # Create grid discretization
        dis = self.get_gwf().dis
        dis_prt = fp.mf6.ModflowGwfdis(prt, nlay=dis.nlay.array, nrow=dis.nrow.array, ncol=dis.ncol.array,
                                        delr=dis.delr.array, delc=dis.delc.array, top=dis.top.array, botm=dis.botm.array,
                                        idomain=dis.idomain.array)

        # create prt input package
        mip = fp.mf6.ModflowPrtmip(prt, pname="mip", porosity=0.2)

        from shapely.geometry import Point, MultiPoint

        grid = self.get_gwf().modelgrid
        ix = fp.utils.gridintersect.GridIntersect(mfgrid=grid)

        list_cellids = []
        for pi in list_p_coords:
            p1 = Point(pi)
            result = ix.intersect(p1)
            if len(result) > 0:
                list_cellids.append(result.cellids[0])
            else:
                p1 = Point((pi[0], pi[1]))
                result = ix.intersect(p1)
                list_cellids.append((-1, result.cellids[0][0], result.cellids[0][1]))

        cellids = np.array([cids for cids in list_cellids])

        cellids[:, 0] += 1

        # package data (irptno, cellid, x, y, z)
        package_data = []
        for i in range(len(cellids)):
            package_data.append((i, cellids[i], list_p_coords[i][0], list_p_coords[i][1], list_p_coords[i][2]))

        # period data (when to release the particles)
        period_data = {0: ["FIRST"]}

        prp = fp.mf6.ModflowPrtprp(prt, pname="prp", filename="{}.prp".format(prt_name),
                                        packagedata=package_data,
                                        perioddata=period_data,
                                        nreleasepts=len(package_data),
                                        exit_solve_tolerance=1e-5)


        # output control package
        budgetfile_prt_name = "{}.bud".format(prt_name)
        trackfile_name = "{}.trk".format(prt_name)
        trackcsvfile_name = "{}.trk.csv".format(prt_name)
        budget_record = [budgetfile_prt_name]
        track_record = [trackfile_name]
        trackcsv_record = [trackcsvfile_name]
        oc = fp.mf6.ModflowPrtoc(
            prt,
            pname="oc",
            budget_filerecord=budget_record,
            track_filerecord=track_record,
            trackcsv_filerecord=trackcsv_record,
            saverecord=[("BUDGET", "ALL")])

        # load head and budget files
        gwf_ws = self.get_gwf().simulation_data.mfpath.get_sim_path()
        head_file_path = os.path.join(gwf_ws, "{}.hds".format(self.model_name))
        head_file = fp.utils.HeadFile(head_file_path, tdis=sim_prt.tdis)
        budget_file_path = os.path.join(gwf_ws, "{}.cbc".format(self.model_name))
        budget_file = fp.utils.CellBudgetFile(budget_file_path, precision="double", tdis=sim_prt.tdis)

        if trackdir == "backward":
            headfile_bkwd_name = f"{self.model_name}_hds_bkwd"
            budgetfile_bkwd_name = f"{self.model_name}_cbc_bkwd"

            # reverse head and budget files to get backward tracking
            head_file.reverse(workspace + "/" + headfile_bkwd_name)
            budget_file.reverse(workspace + "/" + budgetfile_bkwd_name)

            fmi = fp.mf6.ModflowPrtfmi(prt, packagedata=[
                ("GWFHEAD", headfile_bkwd_name),
                ("GWFBUDGET", budgetfile_bkwd_name),
            ])

        else:
            fmi = fp.mf6.ModflowPrtfmi(prt, packagedata=[
                ("GWFHEAD", head_file_path),
                ("GWFBUDGET", budget_file_path),
            ])

        ems = fp.mf6.ModflowEms(sim_prt, pname="ems", filename="{}.ems".format(prt_name))
        sim_prt.register_solution_package(ems, [prt.name])

        self.sim_prt = sim_prt

    def set_porosity(self, 
                     iu = 0, ifa = 0, ip = 0, 
                     porosity=None, k_key="porosity"):

        """
        Set the porosity of the model

        Parameters
        ----------
        iu : int
            unit simulation index
        ifa : int
            facies simulation index
        ip : int
            property simulation index
        porosity : float
            porosity value, if None, the porosity is taken from the table according to the k_key
        k_key : str
            key of the property in the table
        """
        
        gwf = self.get_gwf()
        if porosity is None:
            
            # check if the porosity is already in the table
            assert k_key in self.T1.get_prop_names(), "The property {} is not in the table".format(k_key)

            grid_mode = self.grid_mode
            if grid_mode == "archpy":

                k = self.T1.get_prop(k_key)[iu, ifa, ip]
                k = np.flip(np.flipud(k), axis=1)  # flip the array to have the same orientation as the ArchPy table
                new_k = k

            elif grid_mode == "layers":

                nrow, ncol, nlay = gwf.modelgrid.nrow, gwf.modelgrid.ncol, gwf.modelgrid.nlay

                kh = self.T1.get_prop(k_key)[iu, ifa, ip] 
                new_k = np.ones((nlay, nrow, ncol))

                layers = self.layers_names
                mask_units = [self.T1.unit_mask(l).astype(bool) for l in layers]

                for irow in range(nrow):
                    for icol in range(ncol):
                        for ilay in range(nlay):
                            mask_unit = mask_units[ilay]
                            new_k[ilay, irow, icol] = np.mean(kh[:, irow, icol][mask_unit[:, irow, icol]])

                # fill nan values with the mean of the layer
                for ilay in range(nlay):
                    mask = np.isnan(new_k[ilay])
                    new_k[ilay][mask] = np.nanmean(new_k[ilay])

                new_k = np.flip(new_k, axis=1)  # we have to flip in order to match modflow grid

            elif grid_mode == "new_resolution":
                pass
            
        else:
            new_k = porosity

        mpbas = fp.modpath.Modpath7Bas(self.get_mp(), porosity=new_k)

    def prt_run(self, silent=False):
        # write simulation to disk
        self.sim_prt.write_simulation()
        success, msg = self.sim_prt.run_simulation(silent=silent)
        if not success:
            print("particle tracking did not run successfully")
            print(msg)

    def mp_run(self, silent=False):
        self.mp.write_input()
        success, msg = self.mp.run_model(silent=silent)
        if not success:
            print("modpath did not run successfully")
            print(msg)

    # get results
    def get_mp(self):
        return self.mp
    
    def mp_get_pathlines_object(self):

        fpth = os.path.join(self.model_dir, f"{self.mpnamf}.mppth")
        p = fp.utils.PathlineFile(fpth)
        return p

    def mp_get_endpoints_object(self):

        fpth = os.path.join(self.model_dir, f"{self.mpnamf}.mpend")
        e = fp.utils.EndpointFile(fpth)
        return e

    def mp_get_facies_path_particle(self, i_particle, fac_time = 1/86400, iu = 0, ifa = 0):

        grid_mode = self.grid_mode
        p = self.mp_get_pathlines_object()
        pathline = p.get_data(i_particle)
        df = pd.DataFrame(pathline)
        cells_path = np.array((((df["z"].values-self.T1.zg[0])//self.T1.sz).astype(int),
                                ((df["y"].values-self.T1.yg[0])//self.T1.sy).astype(int),
                                ((df["x"].values-self.T1.xg[0])//self.T1.sx).astype(int))).T
        

        time_ordered = df["time"].values.copy()
        time_ordered *= fac_time
        dt = np.diff(time_ordered)

        # add a column to track distance traveled
        df["distance"] = 0
        for i in range(1, df.shape[0]):
            x0, y0, z0 = df[["x", "y", "z"]].iloc[i-1]
            x1, y1, z1 = df[["x", "y", "z"]].iloc[i]
            distance = np.sqrt((x1-x0)**2 + (y1-y0)**2 + (z1-z0)**2)
            df.loc[df.index[i], "distance"] = distance

        df["cum_distance"] = df["distance"].cumsum()

        # store everything in a new dataframe
        df_all = pd.DataFrame(columns=["dt", "time", "distance", "cum_distance", "x", "y", "z"])
        df_all["dt"] = dt
        df_all["time"] = time_ordered[:-1]
        df_all["distance"] = df["distance"].values[:-1]
        df_all["cum_distance"] = df["cum_distance"].values[:-1]
        df_all["x"] = df["x"].values[:-1]
        df_all["y"] = df["y"].values[:-1]
        df_all["z"] = df["z"].values[:-1]

        if grid_mode in ["layers", "new_resolution"]:
            dic_facies_path = {}
            # retrieve lithologies along the pathlines
            for fa in self.T1.get_all_facies():
                id_fa = fa.ID
                prop_fa = self.upscaled_facies[id_fa]

                facies_along_path = prop_fa[cells_path[:, 0], cells_path[:, 1], cells_path[:, 2]]
                dic_facies_path[fa.ID] = facies_along_path
            colors_fa = []
            for k, v in dic_facies_path.items():
                df_all["facies_prop_"+ str(k)] = v[:-1]
                colors_fa.append(self.T1.get_facies_obj(ID=k, type="ID").c)

        elif grid_mode == "archpy":
            facies = self.T1.get_facies(iu, ifa, all_data=False)
            facies_along_path = facies[cells_path[:, 0], cells_path[:, 1], cells_path[:, 2]]
            df_all["facies"] = facies_along_path[:-1]

        else:
            raise ValueError
        return df_all

    def prt_get_pathlines(self, i_particle=None):
        sim_dir = self.sim_prt.sim_path
        csv_name = self.sim_prt.prt[0].oc.trackcsv_filerecord.array[0][0]
        csv_path = os.path.join(sim_dir, csv_name)
        df = pd.read_csv(csv_path)

        if i_particle is not None:
            df = df.loc[df["irpt"] == i_particle]
        return df

    def prt_get_facies_path_particle(self, i_particle=1, fac_time = 1/86400, iu = 0, ifa = 0):

        grid_mode = self.grid_mode
        df = self.prt_get_pathlines(i_particle)

        time_ordered = df["t"].values.copy()
        time_ordered *= 1/86400
        dt = np.diff(time_ordered)

        # add a column to track distance traveled
        distances = ((df[["x", "y", "z"]].iloc[1:].values - df[["x", "y", "z"]].iloc[:-1].values)**2).sum(1)**0.5

        # store everything in a new dataframe
        df_all = pd.DataFrame(columns=["dt", "time", "distance", "cum_distance", "x", "y", "z"])
        df_all["dt"] = dt
        df_all["time"] = time_ordered[1:]
        df_all["distance"] = distances
        df_all["cum_distance"] = df_all["distance"].cumsum()
        df_all["x"] = (df["x"].values[1:] + df["x"].values[:-1]) / 2
        df_all["y"] = (df["y"].values[1:] + df["y"].values[:-1]) / 2
        df_all["z"] = (df["z"].values[1:] + df["z"].values[:-1]) / 2

        cells_path = np.array((((df_all["z"].values-self.T1.zg[0])//self.T1.sz).astype(int),
                                ((df_all["y"].values-self.T1.yg[0])//self.T1.sy).astype(int),
                                ((df_all["x"].values-self.T1.xg[0])//self.T1.sx).astype(int))).T

        if grid_mode in ["layers", "new_resolution"]:
            dic_facies_path = {}
            # retrieve lithologies along the pathlines
            for fa in self.T1.get_all_facies():
                id_fa = fa.ID
                prop_fa = self.upscaled_facies[id_fa]

                facies_along_path = prop_fa[cells_path[:, 0], cells_path[:, 1], cells_path[:, 2]]
                dic_facies_path[fa.ID] = facies_along_path
            colors_fa = []
            for k, v in dic_facies_path.items():
                df_all["facies_prop_"+ str(k)] = v
                colors_fa.append(self.T1.get_facies_obj(ID=k, type="ID").c)

        elif grid_mode == "archpy":
            facies = self.T1.get_facies(iu, ifa, all_data=False)
            facies_along_path = facies[cells_path[:, 0], cells_path[:, 1], cells_path[:, 2]]
            df_all["facies"] = facies_along_path
        else:
            raise ValueError
        
        return df_all





# Python module to upscale hydraulic conductivity fields
import numpy as np
import matplotlib.pyplot as plt
import geone
import geone.covModel as gcm
from numba import jit

## 2D algorithms ##
# TO DO

## 3D algorithms ##
# Simplified renormalization algorithm

# mean and harmonic mean functions
@jit()
def hmean(v1, v2, w1=1, w2=1):
    return (w1 + w2) / (w1 / v1 + w2 / v2)

@jit()
def mean(v1, v2, w1=1, w2=1):
    return (v1 * w1 + v2 * w2) / (w1 + w2)

def gmean(a):
    if np.all(a > 0):
        return np.exp(np.mean(np.log(a)))
    elif np.all(a < 0):
        a = np.abs(a)
        return -np.exp(np.mean(np.log(a)))
    else:
        raise ValueError("Array must be all positive or all negative")

def fill_nan_values_with_gmean(a):
    
    """
    Function to fill nan values with the geometric mean of the non-nan values
    """

    a = np.copy(a)
    mask = np.isnan(a)
    a[mask] = gmean(a[~mask])
    return a

# function to operate one mean operation
def f_3D(a, dx=None, dy=None, dz=None, direction="x", typ="x"):

    """
    Function to operate a mean between two 2D fields from a 3D field

    Parameters
    ----------
    a : 3D numpy array
        The 3D field
    direction : str
        The direction of the mean to compute. Can be "x", "y" or "z"
    type : str
        The direction where fields are selected. Can be "x", "y" or "z"
    """

    # initialize weights
    w1 = 1
    w2 = 1
    new_dx = None
    new_dy = None
    new_dz = None

    if typ == "y":

        v1 = a[:, ::2, :]
        v2 = a[:, 1::2, :]
        if dy is not None:
            w1 = dy[:, ::2, :]
            w2 = dy[:, 1::2, :]
            new_dy = w1 + w2
            new_dx = dx[:, ::2, :]
            new_dz = dz[:, ::2, :]

    elif typ == "x":

        v1 = a[:, :, ::2]
        v2 = a[:, :, 1::2]
        if dx is not None:
            w1 = dx[:, :, ::2]
            w2 = dx[:, :, 1::2]
            new_dx = w1 + w2
            new_dy = dy[:, :, ::2]
            new_dz = dz[:, :, ::2]

    elif typ == "z":
            
        v1 = a[::2, :, :]
        v2 = a[1::2, :, :]
        if dz is not None:
            w1 = dz[::2, :, :]
            w2 = dz[1::2, :, :]
            new_dz = w1 + w2
            new_dx = dx[::2, :, :]
            new_dy = dy[::2, :, :]


    if typ == direction:  # series
        res = hmean(v1, v2, w1=w1, w2=w2)
    else:
        res = mean(v1, v2, w1=w1, w2=w2)
    
    return res, new_dx, new_dy, new_dz

def test_f():

    """
    Test the function f
    """
    a = np.array([[[0.51394334, 0.77316505],
                   [0.87042769, 0.00804695]],
                  [[0.30973593, 0.95760374],
                   [0.51311671, 0.31828442]]])

    a1 = f_3D(a, direction="x", typ="x")[0]
    a2 = f_3D(a1, direction="x", typ="z")[0]
    a3 = f_3D(a2, direction="x", typ="y")[0]
    res = a3[0, 0, 0]
    assert np.allclose(res, 0.3735857397539791)
    
    a1 = f_3D(a, direction="x", typ="y")[0]
    a2 = f_3D(a1, direction="x", typ="z")[0]
    a3 = f_3D(a2, direction="x", typ="x")[0]
    res = a3[0, 0, 0]
    assert np.allclose(res, 0.5323798506575812)

    print("Test passed")

def merge(v1, v2, w1, w2, normalize=True):
    
    """
    Function to merge two vectors v1 and v2 with weights w1 and w2
    Is used here to merge two columns of different sizes
    """

    if normalize:
        sum_w = w1 + w2
        w1 = w1 / sum_w
        w2 = w2 / sum_w

    return w1*v1 + w2*v2 - (w1*w2*(v1 - v2)**2)/(w1*v2 + w2*v1)


def find_c_3D(k_field, direction="x", typ="min", dx=None, dy=None, dz=None):
    
    if typ == "min":
        if direction == "x":
            l_typ = ["x", "y", "z"]
        elif direction == "y":
            l_typ = ["y", "x", "z"]
        elif direction == "z":
            l_typ = ["z", "x", "y"]
    elif typ == "max":
        if direction == "x":
            l_typ = ["y", "z", "x"]
        elif direction == "y":
            l_typ = ["x", "z", "y"]
        elif direction == "z":
            l_typ = ["x", "y", "z"]

    # if dx, dy and dz are None, set them to 1
    if dx is None:
        dx = 1
    if dy is None:
        dy = 1
    if dz is None:
        dz = 1

    # if dx and dy are provided as a scalar, set them to an array of the same shape as k_field
    if isinstance(dx, (int, float)):
        dx = np.ones(k_field.shape) * dx
    if isinstance(dy, (int, float)):
        dy = np.ones(k_field.shape) * dy
    if isinstance(dz, (int, float)):
        dz = np.ones(k_field.shape) * dz

    a = np.copy(k_field)
    o = 0
    while a.flatten().shape[0] > 1:
        # print("dx:", dx)
        # print("dy:", dy)
        # print("dz:", dz)
        # print("asdf", a)
        # check if the number of elements is odd
        rm_col = False
        rm_row = False
        rm_lay = False
        rm_dx = None
        rm_dy = None
        rm_dz = None
        sim = True
        if l_typ[o] == "x":
            if a.shape[2] > 1:
                if a.shape[2] % 2 == 1:
                    # remove the last column
                    rm_col = a[:, :, -1]
                    a = a[:, :, :-1]
                    rm_dx = dx[:, :, -1]
                    dx = dx[:, :, :-1]
                    rm_dy = dy[:, :, -1]
                    dy = dy[:, :, :-1]
                    rm_dz = dz[:, :, -1]
                    dz = dz[:, :, :-1]
            else:
                sim = False

        elif l_typ[o] == "y":
            if a.shape[1] > 1:
                if a.shape[1] % 2 == 1:
                    # remove the last row
                    rm_row = a[:, -1, :]
                    a = a[:, :-1, :]
                    rm_dx = dx[:, -1, :]
                    dx = dx[:, :-1, :]
                    rm_dy = dy[:, -1, :]
                    dy = dy[:, :-1, :]
                    rm_dz = dz[:, -1, :]
                    dz = dz[:, :-1, :]
            else:
                sim = False

        elif l_typ[o] == "z":
            if a.shape[0] > 1:
                if a.shape[0] % 2 == 1:
                    # remove the last layer
                    rm_lay = a[-1, :, :]
                    a = a[:-1, :, :]
                    rm_dx = dx[-1, :, :]
                    dx = dx[:-1, :, :]
                    rm_dy = dy[-1, :, :]
                    dy = dy[:-1, :, :]
                    rm_dz = dz[-1, :, :]
                    dz = dz[:-1, :, :]
            else:
                sim = False
        if sim:
            a, dx, dy, dz = f_3D(a, direction=direction, typ=l_typ[o], dx=dx, dy=dy, dz=dz)

        # merge the removed column, row or layer
        if rm_col is not False:
            # new_col = merge(rm_col, a[:, :, -1], 1, 2)
            a = np.dstack((a, rm_col.reshape(a.shape[0], a.shape[1], 1)))
            dx = np.dstack((dx, rm_dx.reshape(dx.shape[0], dx.shape[1], 1)))
            dy = np.dstack((dy, rm_dy.reshape(dy.shape[0], dy.shape[1], 1)))
            dz = np.dstack((dz, rm_dz.reshape(dz.shape[0], dz.shape[1], 1)))
        elif rm_row is not False:
            # new_row = merge(rm_row, a[:, -1, :], 1, 2)
            a = np.hstack((a, rm_row.reshape(a.shape[0], 1, a.shape[2])))
            dx = np.hstack((dx, rm_dx.reshape(dx.shape[0], 1, dx.shape[2])))
            dy = np.hstack((dy, rm_dy.reshape(dy.shape[0], 1, dy.shape[2])))
            dz = np.hstack((dz, rm_dz.reshape(dz.shape[0], 1, dz.shape[2])))
        elif rm_lay is not False:
            # new_lay = merge(rm_lay, a[-1, :, :], 1, 2)
            a = np.vstack((a, rm_lay.reshape(1, a.shape[1], a.shape[2])))
            dx = np.vstack((dx, rm_dx.reshape(1, dx.shape[1], dx.shape[2])))
            dy = np.vstack((dy, rm_dy.reshape(1, dy.shape[1], dy.shape[2])))
            dz = np.vstack((dz, rm_dz.reshape(1, dz.shape[1], dz.shape[2])))

        o += 1
        if o == 3:
            o = 0

    return a[0, 0, 0]

def test_find_c():

    b = np.array([[[0.51394334, 0.77316505, 0.87042769],
        [0.00804695, 0.30973593, 0.95760374],
        [0.51311671, 0.31828442, 0.53919994]],

       [[0.22125494, 0.80648136, 0.34225463],
        [0.53888885, 0.00587379, 0.67315248],
        [0.21002426, 0.93255759, 0.37424475]],

       [[0.75241892, 0.763139  , 0.87049946],
        [0.11011118, 0.30009198, 0.47490577],
        [0.67293672, 0.25759243, 0.70115132]]])

    cmin = find_c_3D(b, direction="x", typ="min", dx=1, dy=1, dz=1)
    cmax = find_c_3D(b, direction="x", typ="max", dx=1, dy=1, dz=1)
    assert np.allclose(cmin, 0.3914114)
    assert np.allclose(cmax, 0.47437664)

    print("Test passed")

# now keq is a function of cmin and cmax according to keq = cmax**alpha*(cmin)**(1-alpha)
def get_alpha(dx, dy, dz=None, direction="x"):
    
    def f_u(t):
        return np.arctan(np.sqrt(t)) / (np.pi/2)

    if dz is None:
        if direction == "x":
            a = dx/dy
        elif direction == "y":
            a = dy/dx
        
        u = f_u(a)
        return u

    if direction == "x":
        a_1 = dx/dy
        a_2 = dx/dz
    elif direction == "y":
        a_1 = dy/dx
        a_2 = dy/dz
    elif direction == "z":
        a_1 = dz/dx
        a_2 = dz/dy
        
    u_1 = f_u(a_1)
    u_2 = f_u(a_2)

    return ((u_1 + u_2) - 2*u_1*u_2) / (1 - u_1*u_2)
    
def keq(cmin, cmax, alpha):

    res = (np.abs(cmax)**alpha)*np.abs(cmin)**(1-alpha)
    if cmax and cmin < 0:
        res = -res

    return res

def simplified_renormalization(field, dx, dy, dz, direction="x"):

    cx_min, cx_max = find_c_3D(field, direction=direction, typ="min", dx=dx, dy=dy, dz=dz), find_c_3D(field, direction=direction, typ="max", dx=dx, dy=dy, dz=dz)
    return keq(cx_min, cx_max, get_alpha(dx, dy, dz, direction=direction)) 

# Standard renormalization algorithm
def get_idx (i):
    
    if i == 1:
        idx = (0, 0, 0)
    elif i == 2:
        idx = (0, 0, 1)
    elif i == 3:
        idx = (0, 1, 0)
    elif i == 4:
        idx = (0, 1, 1)
    elif i == 5:
        idx = (1, 0, 0)
    elif i == 6:
        idx = (1, 0, 1)
    elif i == 7:
        idx = (1, 1, 0)
    elif i == 8:
        idx = (1, 1, 1)
    
    return idx

def f_sr(kxx, kyy, kzz, dx, dy, dz, direction="x"):

    if direction == "x":    
        pass
    elif direction == "y":
        # rotate the fields
        kxx = np.rot90(kxx, k=-1, axes=(1, 2))
        kyy = np.rot90(kyy, k=-1, axes=(1, 2))
        kzz = np.rot90(kzz, k=-1, axes=(1, 2))
        # adjust the dx and dy
        dx, dy = dy, dx
    elif direction == "z":
        # rotate the fields
        kxx = np.rot90(kxx, k=-1, axes=(0, 2))
        kyy = np.rot90(kyy, k=-1, axes=(0, 2))
        kzz = np.rot90(kzz, k=-1, axes=(0, 2))
        # adjust the dx and dz
        dx, dz = dz, dx
        
    def b1(i, u):

        idx = get_idx(i)

        if u == "x":
            du = dx
            ku = kxx[idx]
        elif u == "y":
            du = dy
            ku = kyy[idx]
        elif u == "z":
            du = dz
            ku = kzz[idx]

        return 2 * ku * (dx*dy*dz) / du**2

    def b2(i, j, u):
            
        idx = get_idx(i)
    
        idx2 = get_idx(j)
    
        if u == "x":
            du = dx
            ku = kxx[idx]
            ku2 = kxx[idx2]
        elif u == "y":
            du = dy
            ku = kyy[idx]
            ku2 = kyy[idx2]
        elif u == "z":
            du = dz
            ku = kzz[idx]
            ku2 = kzz[idx2]
    
        return ((2 * ku * ku2) /(ku + ku2)) * ((dx*dy*dz) / du**2)

    def d(i, j, k, l):
        return b1(i, "x") + b2(i, j, "x") + b2(i, k, "y") + b2(i, l, "z")


    A = np.array(
        [
            [d(1, 2, 3, 5), -b2(1, 2, "x"), -b2(1, 3, "y"), 0, -b2(1, 5, "z"), 0, 0, 0],
            [-b2(2, 1, "x"), d(2, 1, 4, 6), 0, -b2(2, 4, "y"), 0, -b2(2, 6, "z"), 0, 0],
            [-b2(3, 1, "y"), 0, d(3, 4, 1, 7), -b2(3, 4, "x"), 0, 0, -b2(3, 7, "z"), 0],
            [0, -b2(4, 2, "y"), -b2(4, 3, "x"), d(4, 3, 2, 8), 0, 0, 0, -b2(4, 8, "z")],
            [-b2(5, 1, "z"), 0, 0, 0, d(5, 6, 7, 1), -b2(5, 6, "x"), -b2(5, 7, "y"), 0],
            [0, -b2(6, 2, "z"), 0, 0, -b2(6, 5, "x"), d(6, 5, 8, 2), 0, -b2(6, 8, "y")],
            [0, 0, -b2(7, 3, "z"), 0, -b2(7, 5, "y"), 0, d(7, 8, 5, 3), -b2(7, 8, "x")],
            [0, 0, 0, -b2(8, 4, "z"), 0, -b2(8, 6, "y"), -b2(8, 7, "x"), d(8, 7, 6, 4)]
        ])
    
    b = np.array([b1(1, "x"), 0, b1(3, "x"), 0, b1(5, "x"), 0, b1(7, "x"), 0])	

    h = np.linalg.solve(A, b)
    K = (h[1::2] * kxx[:, :, -1].flatten()).sum()

    return K

# standard renormalization with direct scheme
def f_sr_center(kxx, kyy, kzz, dx, dy, dz):

    def b1(i, u):

        idx = get_idx(i)

        if u == "x":
            du = dx
            ku = kxx[idx]
        elif u == "y":
            du = dy
            ku = kyy[idx]
        elif u == "z":
            du = dz
            ku = kzz[idx]

        return 2 * ku * (dx*dy*dz) / du**2

    def b2(i, j, u):
            
        idx = get_idx(i)
    
        idx2 = get_idx(j)
    
        if u == "x":
            du = dx
            ku = kxx[idx]
            ku2 = kxx[idx2]
        elif u == "y":
            du = dy
            ku = kyy[idx]
            ku2 = kyy[idx2]
        elif u == "z":
            du = dz
            ku = kzz[idx]
            ku2 = kzz[idx2]
    
        return (2 * ku * ku2) /(ku + ku2) * ((dx*dy*dz) / du**2)

    # Kxx
    def d(i, j, k, l):
        return b1(i, "x") + b1(j, "x") + b1(k, "y") + b1(l, "z")

    A = np.array([[d(2, 1, 2, 2), -b1(2, "y"), -b1(2, "z"), 0],
                  [-b1(2, "y"), d(4, 3, 2, 4), 0, -b1(4, "z")],
                  [-b1(2, "z"), 0, d(6, 5, 6, 2), -b1(6, "y")],
                  [0, -b1(4, "z"), -b1(6, "y"), d(8, 7, 6, 4)]])

    b = np.array([b1(1, "x"), b1(3, "x"), b1(5, "x"), b1(7, "x")])

    h = np.linalg.solve(A, b)
    Kxx = 0.5 * (h * kxx[:, :, -1].flatten()).sum()

    # Kyy
    def d(i, j, k, l):
        return b1(i, "y") + b1(j, "x") + b1(k, "y") + b1(l, "z")
    
    A = np.array([[d(3, 3, 1, 3), -b1(3, "x"), -b1(3, "z"), 0],
                    [-b1(3, "x"), d(2, 3, 2, 4), 0, -b1(4, "z")],
                    [-b1(3, "z"), 0, d(7, 7, 5, 3), -b1(7, "x")],
                    [0, -b1(4, "z"), -b1(7, "x"), d(8, 7, 6, 4)]])

    b = np.array([b1(1, "y"), b1(2, "y"), b1(5, "y"), b1(6, "y")])

    h = np.linalg.solve(A, b)
    Kyy = 0.5 * (h * kyy[:, -1, :].flatten()).sum()

    # Kzz
    def d(i, j, k, l):
        return b1(i, "z") + b1(j, "x") + b1(k, "y") + b1(l, "z")

    A = np.array([[d(5, 5, 5, 1), -b1(5, "x"), -b1(5, "y"), 0],
                    [-b1(5, "x"), d(6, 5, 6, 2), 0, -b1(6, "y")],
                    [-b1(5, "y"), 0, d(7, 7, 5, 3), -b1(7, "x")],
                    [0, -b1(6, "y"), -b1(7, "x"), d(8, 7, 6, 4)]])

    b = np.array([b1(1, "z"), b1(2, "z"), b1(3, "z"), b1(4, "z")])

    h = np.linalg.solve(A, b)
    Kzz = 0.5 * (h * kzz[-1, :, :].flatten()).sum()

    return Kxx, Kyy, Kzz

def standard_renormalization(field_xx, field_yy, field_zz, dx, dy, dz, niter=1, scheme="direct"):
    
    assert field_xx.shape == field_yy.shape == field_zz.shape, "Field shapes must be the same"
    assert np.log2(field_xx.shape[0]).is_integer(), "Field size must be a power of 2"
    assert np.log2(field_xx.shape[1]).is_integer(), "Field size must be a power of 2"
    assert np.log2(field_xx.shape[2]).is_integer(), "Field size must be a power of 2"

    def one_iteration(field_xx, field_yy, field_zz):
        new_field_xx = np.zeros((field_xx.shape[0]//2, field_xx.shape[1]//2, field_xx.shape[2]//2))
        new_field_yy = np.zeros((field_xx.shape[0]//2, field_xx.shape[1]//2, field_xx.shape[2]//2))
        new_field_zz = np.zeros((field_xx.shape[0]//2, field_xx.shape[1]//2, field_xx.shape[2]//2))
        for i in range(0, field_xx.shape[0], 2):
            for j in range(0, field_xx.shape[1], 2):
                for k in range(0, field_xx.shape[2], 2):
                    block_xx = field_xx[i:i+2, j:j+2, k:k+2]
                    block_yy = field_yy[i:i+2, j:j+2, k:k+2]
                    block_zz = field_zz[i:i+2, j:j+2, k:k+2]

                    # fill nan values with the geometric mean
                    block_xx = fill_nan_values_with_gmean(block_xx)
                    block_yy = fill_nan_values_with_gmean(block_yy)
                    block_zz = fill_nan_values_with_gmean(block_zz)

                    if scheme == "direct":
                        kxx = f_sr(block_xx, block_yy, block_zz, dx, dy, dz, direction="x")
                        kyy = f_sr(block_yy, block_xx, block_zz, dy, dx, dz, direction="y")
                        kzz = f_sr(block_zz, block_xx, block_yy, dz, dx, dy, direction="z")
                    elif scheme == "center":
                        kxx, kyy, kzz = f_sr_center(block_xx, block_yy, block_zz, dx, dy, dz)
                    
                    new_field_xx[i//2, j//2, k//2] = kxx
                    new_field_yy[i//2, j//2, k//2] = kyy
                    new_field_zz[i//2, j//2, k//2] = kzz

        return new_field_xx, new_field_yy, new_field_zz

    # list_field_xx = [field_xx]
    # list_field_yy = [field_yy]
    # list_field_zz = [field_zz]

    for _ in range(niter):
        field_xx, field_yy, field_zz = one_iteration(field_xx, field_yy, field_zz)
        # list_field_xx.append(field_xx)
        # list_field_yy.append(field_yy)
        # list_field_zz.append(field_zz)

    return field_xx, field_yy, field_zz

# Tensorial renormalization algorithm
def f_rt_simple(kxx, kyy, kzz, dx, dy, dz):
    
    def b(i, u, v):

        idx = get_idx(i)
        if u != v:
            return 0

        elif u == "x" and v == "x":
            kuv = kxx
            du = dx
            dv = dx
        elif u == "y" and v == "y":
            kuv = kyy
            du = dy
            dv = dy
        elif u == "z" and v == "z":
            kuv = kzz
            du = dz
            dv = dz
        else:
            raise ValueError("Invalid indices")
        
        return kuv[idx] * (dx*dy*dz) / (du*dv)
    
    
    def d(i, j, k, l):
        return b(i, "x", "x") + b(i, "y", "y") + b(i, "z", "z") + b(j, "x", "x") + b(k, "y", "y") + b(l, "z", "z")

    def m1(i, j, u):
        return -b(i, u, u) - b(j, u, u)
    

    A = np.array(
        [
            [d(1, 2 ,3 ,5), m1(1, 2, "x"), m1(1, 3, "y"), 0, m1(1, 5, "z"), 0, 0],
            [m1(2, 1, "x"), d(2, 1, 4, 6), 0, m1(2, 4, "y"), 0, m1(2, 6, "z"), 0],
            [m1(3, 1, "y"), 0, d(3, 4, 1, 7), m1(3, 4, "x"), 0, 0, m1(3, 7, "z")],
            [0, m1(4, 2, "y"), m1(4, 3, "x"), d(4, 3, 2, 8), 0, 0, 0],
            [m1(5, 1, "z"), 0, 0, 0, d(5, 6, 7, 1), m1(5, 6, "x"), m1(5, 7, "y")],
            [0, m1(6, 2, "z"), 0, 0, m1(6, 5, "x"), d(6, 5, 8, 2), 0],
            [0, 0, m1(7, 3, "z"), 0, m1(7, 5, "y"), 0, d(7, 8, 5, 3)]
        ]
        )
    
    B = np.array(
        [
            [ b(2, "x", "x"),  b(3, "y", "y"),  b(5, "z", "z")],
            [-b(2, "x", "x"),  b(4, "y", "y"),  b(6, "z", "z")],
            [ b(4, "x", "x"), -b(3, "y", "y"),  b(7, "z", "z")],
            [-b(4, "x", "x"),- b(4, "y", "y"),  b(8, "z", "z")],
            [ b(6, "x", "x"),  b(7, "y", "y"), -b(5, "z", "z")],
            [-b(6, "x", "x"),  b(8, "y", "y"), -b(6, "z", "z")],
            [ b(8, "x", "x"), -b(7, "y", "y"), -b(7, "z", "z")]
        ]
        )

    T = np.array(
        [
            [b(2, "x", "x") + b(4, "x", "x") + b(6, "x", "x") + b(8, "x", "x"), 0, 0],
            [0, b(3, "y", "y") + b(4, "y", "y") + b(7, "y", "y") + b(8, "y", "y"), 0],
            [0, 0, b(5, "z", "z") + b(6, "z", "z") + b(7, "z", "z") + b(8, "z", "z")]    
        ])
    
    # solve the system of equations
    P = np.linalg.solve(A, B)
    dxdydz = np.array([[dx**2, 0, 0], [0, dy**2, 0], [0, 0, dz**2]])
    K = 0.5 * ((-np.dot(B.T, P) + T) @ dxdydz) * (1/dx * 1/dy * 1/dz) 
    return K

def tensorial_renormalization(field_xx, field_yy, field_zz, dx, dy, dz, niter=1):
    
    # assert that field is 2**n
    assert field_xx.shape == field_yy.shape == field_zz.shape, "Field shapes must be the same"
    # assert field_xx.shape[0] == field_xx.shape[1] == field_xx.shape[2]
    assert np.log2(field_xx.shape[0]).is_integer(), "Field size must be a power of 2"
    assert np.log2(field_xx.shape[1]).is_integer(), "Field size must be a power of 2"
    assert np.log2(field_xx.shape[2]).is_integer(), "Field size must be a power of 2"

    # assert np.log2(field.shape[0]).is_integer(), "Field size must be a power of 2
    
    def one_iteration(field_xx, field_yy, field_zz):
        new_field_xx = np.zeros((field_xx.shape[0]//2, field_xx.shape[1]//2, field_xx.shape[2]//2))
        new_field_yy = np.zeros((field_xx.shape[0]//2, field_xx.shape[1]//2, field_xx.shape[2]//2))
        new_field_zz = np.zeros((field_xx.shape[0]//2, field_xx.shape[1]//2, field_xx.shape[2]//2))
        for i in range(0, field_xx.shape[0], 2):
            for j in range(0, field_xx.shape[1], 2):
                for k in range(0, field_xx.shape[2], 2):
                    block_xx = field_xx[i:i+2, j:j+2, k:k+2]
                    block_yy = field_yy[i:i+2, j:j+2, k:k+2]
                    block_zz = field_zz[i:i+2, j:j+2, k:k+2]

                    # fill nan values with the geometric mean
                    block_xx = fill_nan_values_with_gmean(block_xx)
                    block_yy = fill_nan_values_with_gmean(block_yy)
                    block_zz = fill_nan_values_with_gmean(block_zz)

                    K = f_rt_simple(block_xx, block_yy, block_zz, dx, dy, dz)
                    kxx = K[0, 0]
                    kyy = K[1, 1]
                    kzz = K[2, 2]
                    
                    new_field_xx[i//2, j//2, k//2] = kxx
                    new_field_yy[i//2, j//2, k//2] = kyy
                    new_field_zz[i//2, j//2, k//2] = kzz

        return new_field_xx, new_field_yy, new_field_zz

    # list_field_xx = [field_xx]
    # list_field_yy = [field_yy]
    # list_field_zz = [field_zz]

    for _ in range(niter):
        field_xx, field_yy, field_zz = one_iteration(field_xx, field_yy, field_zz)
        # list_field_xx.append(field_xx)
        # list_field_yy.append(field_yy)
        # list_field_zz.append(field_zz)

    return field_xx, field_yy, field_zz



def upscale_k(field, dx=1, dy=1, dz=1, method="simplified_renormalization", factor_x=2, factor_y=2, factor_z=2, grid=None, scheme="center"):

    """
    Function to upscale a field using renormalization methods and average methods

    Parameters
    ----------
    field : np.ndarray
        3D isotropic hydraulic conductivity field
    dx : float or np.ndarray
        Cell size in the x direction
    dy : float or np.ndarray
        Cell size in the y direction
    dz : float or np.ndarray
        Cell size in the z direction
    method : str
        Renormalization method. Options are simplified_renormalization, standard_renormalization, tensorial_renormalization, arithmetic, harmonic and geometric
    factor_x : int
        Upscaling factor in the x direction. This means that the new field will have a shape divided by factor_x in the x direction
        Must be a power of 2 for standard and tensorial renormalization. It should be noted that field shape in the x direction must be a multiple of factor_x
    factor_y : int
        Upscaling factor in the y direction. This means that the new field will have a shape divided by factor_y in the y direction
        Must be a power of 2 for standard and tensorial renormalization. It should be noted that field shape in the y direction must be a multiple of factor_y
    factor_z : int
        Upscaling factor in the z direction. This means that the new field will have a shape divided by factor_z in the z direction
        Must be a power of 2 for standard and tensorial renormalization. It should be noted that field shape in the z direction must be a multiple of factor_z
    grid : np.ndarray
        Modflow dis grid. Only works with simplified_renormalization, arithmetic, harmonic and geometric. 
        Not implemented yet.
    scheme : str
        Scheme for standard renormalization. Options are direct and center. Center is generally more accurate and faster. Default is center
    """

    assert field.shape[1] % factor_y == 0, "factor_y must be a divisor of the field size"
    assert field.shape[2] % factor_x == 0, "factor_x must be a divisor of the field size"
    assert field.shape[0] % factor_z == 0, "factor_z must be a divisor of the field size"

    if method == "simplified_renormalization":
        
        if grid is None:
            new_field_kxx = np.zeros((field.shape[0]//factor_z, field.shape[1]//factor_y, field.shape[2]//factor_x))
            new_field_kyy = np.zeros((field.shape[0]//factor_z, field.shape[1]//factor_y, field.shape[2]//factor_x))
            new_field_kzz = np.zeros((field.shape[0]//factor_z, field.shape[1]//factor_y, field.shape[2]//factor_x))

            for i in range(0, field.shape[0], factor_z):
                for j in range(0, field.shape[1], factor_y):
                    for k in range(0, field.shape[2], factor_x):
                        selected_area = field[i:i+factor_z, j:j+factor_y, k:k+factor_z]

                        # fill nan values with the geometric mean of the selected area
                        selected_area = fill_nan_values_with_gmean(selected_area)

                        Kxx = simplified_renormalization(selected_area, dx, dy, dz, direction="x")
                        Kyy = simplified_renormalization(selected_area, dx, dy, dz, direction="y")
                        Kzz = simplified_renormalization(selected_area, dx, dy, dz, direction="z")

                        new_field_kxx[i//factor_z, j//factor_y, k//factor_x] = Kxx
                        new_field_kyy[i//factor_z, j//factor_y, k//factor_x] = Kyy
                        new_field_kzz[i//factor_z, j//factor_y, k//factor_x] = Kzz

        return new_field_kxx, new_field_kyy, new_field_kzz

    elif method == "standard_renormalization":
        assert factor_x == factor_y == factor_z, "factor_x, factor_y and factor_z must be equal for standard renormalization"
        assert np.log2(factor_x).is_integer(), "factor_x must be a power of 2"

        field_kxx, field_kyy, field_kzz = standard_renormalization(field, field, field, dx, dy, dz, niter=int(np.log2(factor_x)), scheme=scheme)
        return field_kxx, field_kyy, field_kzz
    
    elif method == "tensorial_renormalization":
        assert factor_x == factor_y == factor_z, "factor_x, factor_y and factor_z must be equal for tensorial renormalization"
        assert np.log2(factor_x).is_integer(), "factor_x must be a power of 2"

        field_kxx, field_kyy, field_kzz = tensorial_renormalization(field, field, field, dx, dy, dz, niter=int(np.log2(factor_x)))
        return field_kxx, field_kyy, field_kzz

    elif method in ["arithmetic", "harmonic", "geometric"]:

        if grid is None:
            new_field = np.zeros((field.shape[0]//factor_z, field.shape[1]//factor_y, field.shape[2]//factor_x))

            for i in range(0, field.shape[0], factor_z):
                for j in range(0, field.shape[1], factor_y):
                    for k in range(0, field.shape[2], factor_x):
                        selected_area = field[i:i+factor_z, j:j+factor_y, k:k+factor_x]

                        # remove nan values
                        selected_area = selected_area[~np.isnan(selected_area)]

                        if method == "arithmetic":
                            K = np.mean(selected_area.flatten())
                        elif method == "harmonic":
                            K = 1 / np.mean(1 / selected_area.flatten())
                        elif method == "geometric":
                            if np.all(selected_area < 0):
                                selected_area = np.abs(selected_area)
                                K = -np.exp(np.mean(np.log(selected_area.flatten())))
                            else:
                                K = np.exp(np.mean(np.log(selected_area.flatten())))
                        new_field[i//factor_z, j//factor_y, k//factor_x] = K

        return new_field

    else:
        raise ValueError("method must be simplified_renormalization, standard_renormalization, tensorial_renormalization, arithmetic, harmonic or geometric")